@@ -1,10 +1,10 @@
-from typing import Dict, Iterable, Optional, Iterator, Union
+from typing import Dict, Iterable, Optional, Union
 
 import torch
-from transformers import DataCollatorForTokenClassification, DataCollatorWithPadding, BatchEncoding
+from transformers import DataCollatorForTokenClassification, BatchEncoding
 
 from ..objectives.objective_base import SupervisedObjective
-from ..utils import AdaptationDataset, Head
+from ..utils import Head
 
 
 class TokenClassification(SupervisedObjective):
@@ -79,19 +79,16 @@ class TokenClassification(SupervisedObjective):
         :return: Encodings compatible with self.compatible_model_head.
         """
 
-        if self.texts is not None:
-            texts_iter = iter(self.texts)
-            labels_iter = iter(self.labels)
-        else:
-            texts_iter = AdaptationDataset.iter_text_file_per_line(self.texts_path)
-            labels_iter = AdaptationDataset.iter_text_file_per_line(self.labels_path)
+        texts_iter, labels_iter = self._per_split_iterators(split)
 
         aligned_collated_iter = self._wordpiece_token_label_alignment(texts_iter, labels_iter)
 
         return aligned_collated_iter
 
-    def _compute_loss(self, logit_outputs: torch.FloatTensor,
+    def _compute_loss(self,
+                      logit_outputs: torch.FloatTensor,
                       labels: torch.LongTensor,
+                      inputs: Optional[Union[BatchEncoding, Dict[str, torch.Tensor]]] = None,
                       attention_mask: Optional[torch.LongTensor] = None) -> torch.FloatTensor:
         """
         Computes a loss for model outputs on a single token classification batch.
@@ -121,40 +118,13 @@ class SequenceClassification(SupervisedObjective):
 
     compatible_head = Head.SEQ_CLASSIFICATION
 
-    def _get_inputs_iterator(self, split: str) -> Iterator:
-        """
-        Batches and encodes input texts and corresponding labels.
-        :param split: Selected data split. `train` or `eval`.
-        :return: Iterator over batch encodings.
-        """
-
-        collator = DataCollatorWithPadding(self.tokenizer, pad_to_multiple_of=8)
-        classifying_pairs = None
-
-        batch_features = []
-        for src_text, label in zip(*self._per_split_iterators(split)):
-            # check from the first sample
-            if classifying_pairs is None:
-                # if the input texts are tab-separated we will tokenize them as pairs
-                classifying_pairs = "\t" in src_text
-            if classifying_pairs:
-                text, text_pair = src_text.split("\t")
-                out_sample = self.tokenizer(text, text_pair=text_pair, truncation=True)
-            else:
-                out_sample = self.tokenizer(src_text, truncation=True)
-            out_sample["label"] = torch.tensor(self.labels_map[label])
-            batch_features.append(out_sample)
-            if len(batch_features) == self.batch_size:
-                yield collator(batch_features)
-                batch_features = []
-
-        if batch_features:
-            # yield residual batch
-            yield collator(batch_features)
-
-    def _compute_loss(self, logit_outputs: torch.FloatTensor, labels: torch.LongTensor) -> torch.FloatTensor:
+    def _compute_loss(self,
+                      logit_outputs: torch.FloatTensor,
+                      labels: torch.LongTensor,
+                      inputs: Optional[Union[BatchEncoding, Dict[str, torch.Tensor]]] = None) -> torch.FloatTensor:
         """
         Computes a loss for model outputs on a single sequence classification batch.
+        :param inputs: Input encoding corresponding to given `logit_outputs` and `labels`.
         :param logit_outputs: Sequence Classification model raw outputs.
         :param labels: Expected labels.
         :return: loss value with grad_fn.
