@@ -1,17 +1,16 @@
 import abc
 import itertools
 import logging
-from typing import List, Union, Optional, Iterable, Tuple, Dict, Sequence, Any
+from typing import List, Union, Optional, Iterable, Tuple, Dict, Sequence, Any, Iterator
 
 import torch
 from tqdm import trange
-from transformers import BatchEncoding
+from transformers import BatchEncoding, DataCollatorWithPadding
 
 from ..evaluators.evaluator_base import EvaluatorBase
 from ..lang_module import LangModule
 from ..utils import AdaptationDataset, Head, TransformerAdaptationDataset
 
-
 logger = logging.getLogger()
 
 
@@ -24,6 +23,7 @@ class Objective(abc.ABC):
     given_id: Optional[str]
     epoch: int
     num_steps: int
+    last_input: Optional[Union[BatchEncoding, Dict[str, torch.Tensor]]]
 
     texts: Optional[List[str]]
     texts_path: Optional[str]
@@ -51,7 +51,8 @@ class Objective(abc.ABC):
                  objective_id: Optional[str] = "",
                  loss_weight: Optional[float] = 1,
                  max_samples_per_log: int = 1000,
-                 max_samples_per_eval_log: int = 10000):
+                 max_samples_per_eval_log: int = 10000,
+                 remember_last_input: Optional[bool] = False):
         """
         Shared initialisation logic of every Objective.
         Registers a compatible model for this objective to given `lang_module`,
@@ -71,6 +72,10 @@ class Objective(abc.ABC):
         :param loss_weight: A scalar of the loss of this objective in multi-objective training.
         :param max_samples_per_log: Maximum number batches that this objective will compute for logging.
         :param max_samples_per_eval_log: Maximum number batches that this objective will compute for evaluation logging.
+        :param remember_last_input: Debugging feature: whether the objective should remember the last input
+        to its compatible model. Useful for debugging a development of the new objective;
+        If the training fails (in the interactive - `-i` mode), the last, possibly error input can be retrieved
+        from `this_objective.last_input`.
         """
 
         self.batch_size = batch_size
@@ -78,6 +83,8 @@ class Objective(abc.ABC):
         self.given_id = objective_id
         self.loss_weight = loss_weight
         self.num_steps = 0
+        self.remember_last_input = remember_last_input
+        self.last_input = None
 
         self.compatible_head_model = self.register_compatible_head_model(lang_module,
                                                                          share_other_objective_head,
@@ -138,10 +145,12 @@ class Objective(abc.ABC):
 
         out_logs["%s_%s_loss" % (split, self)] = mean_loss
         out_logs["%s_%s_num_batches" % (split, self)] = len(loss_history)
+
         for evaluator in self.evaluators[split]:
             dataset = self.get_dataset(split, 0, self.compatible_head_model.device,
-                                       firstn=self.max_samples_per_log[split], add_oid=False)
-
+                                       firstn=self.max_samples_per_log[split],
+                                       add_oid=False,
+                                       is_training_dataset=False)
             # evaluator should already return an aggregated value, so unlike loss, we don't average it
             evaluator_value = evaluator(self.compatible_head_model, self.tokenizer, dataset)
             self.evaluations_history[split][evaluator].append(evaluator_value)
@@ -201,31 +210,40 @@ class Objective(abc.ABC):
         return passed_patience_evals and did_not_improve
 
     @abc.abstractmethod
-    def _compute_loss(self, logit_outputs: torch.FloatTensor, labels: torch.LongTensor) -> torch.FloatTensor:
+    def _compute_loss(self,
+                      logit_outputs: torch.FloatTensor,
+                      labels: torch.LongTensor,
+                      inputs: Optional[Union[BatchEncoding, Dict[str, torch.Tensor]]] = None) -> torch.FloatTensor:
         """
         An implementation of the loss computation for a given objective.
         Override this, or inherit it from other suitable objective when implementing custom objective.
+        :param inputs: Input encoding corresponding to given `logit_outputs` and `labels`.
         :param logit_outputs: Raw output of this objective's head.
         :param labels: Expected true labels of this objective.
         :return: a single-item torch tensor with registered grad_fn.
         """
         pass
 
-    def compute_loss(self, logit_outputs: torch.FloatTensor, labels: torch.LongTensor, split: str) -> torch.FloatTensor:
+    def compute_loss(self,
+                     logit_outputs: torch.FloatTensor,
+                     labels: torch.LongTensor,
+                     inputs: Union[BatchEncoding, Dict[str, torch.Tensor]] = None,
+                     split: Optional[str] = "") -> torch.FloatTensor:
         """
         Shared wrapper of objective-specific loss computation. Additionally, it registers model outputs, and labels
         for logging and updates this objective progress bar.
-        :param logit_outputs:
-        :param labels:
-        :param split:
-        :return:
+        :param inputs: Input encoding corresponding to given `logit_outputs` and `labels`.
+        :param logit_outputs: Raw output of this objective's head.
+        :param labels: Expected true labels of this objective.
+        :param split: Dataset split. `train` or `eval`.
+        :return: a single-item torch tensor with registered grad_fn.
         """
-        loss = self._compute_loss(logit_outputs, labels)
+        loss = self._compute_loss(logit_outputs, labels, inputs)
         self.loss_history[split].append(loss.item())
         self.num_steps += 1
 
-        self.progressbar[split].set_postfix(refresh=False, split=split, loss=loss.item(), epoch=self.epoch)
-        self.progressbar[split].update(1)
+        if self.progressbar[split] is not None:
+            self.progressbar[split].set_postfix(refresh=False, split=split, loss=loss.item(), epoch=self.epoch)
 
         return loss * self.loss_weight
 
@@ -247,7 +265,9 @@ class Objective(abc.ABC):
                     objective_i: int,
                     device: Union[str, torch.device],
                     firstn: Optional[int] = None,
-                    add_oid: bool = True) -> TransformerAdaptationDataset:
+                    add_oid: bool = True,
+                    is_training_dataset: bool = True,
+                    show_progressbar: bool = True) -> TransformerAdaptationDataset:
         """
         Default logic for wrapping the inputs iterator into torch.IterableDataset, used in Trainer.train_dataloaer.
         :param split: A split of the retrieved dataset. `train` or `eval`.
@@ -255,17 +275,25 @@ class Objective(abc.ABC):
         :param device: Device to transfer this data set to.
         :param firstn: If given, a number of the retrieved items from the dataset.
         :param add_oid: Whether to append objective id to the match. Required for forward pass over LangModule.
+        :param is_training_dataset: Whether this dataset is used for training -> if to update the epochs counter.
 
         :return: TransformerAdaptationDataset wrapping a data set of this objective.
         """
-        self.epoch += 1 if split == "train" else 0
+        if split == "train" and is_training_dataset:
+            # increment epoch only for train split and only for the dataset used as training
+            # - get_dataset is also called from self.per_objective_log, or specific objectives
+            self.epoch += 1 if split == "train" else 0
 
-        self.progressbar[split] = trange(self.dataset_length[split] // self.batch_size,
-                                         desc=str(self),
-                                         unit="batches",
-                                         position=objective_i,
-                                         leave=True)
-        self.progressbar[split].set_postfix(refresh=False, split=split, epoch=self.epoch, loss=-1)
+        if show_progressbar:
+            self.progressbar[split] = trange(self.dataset_length[split] // self.batch_size,
+                                             desc=str(self),
+                                             unit="batches",
+                                             position=objective_i,
+                                             leave=True)
+            self.progressbar[split].set_postfix(refresh=False, split=split, epoch=self.epoch, loss=-1)
+        else:
+            # we do not update loss, if no progress bar is pertained
+            self.progressbar[split] = None
 
         inputs_iter = self._get_inputs_iterator(split)
 
@@ -276,6 +304,14 @@ class Objective(abc.ABC):
             sample["oid"] = id(self)
             return sample
 
+        def _remember_input(sample: Union[BatchEncoding, Dict[str, torch.LongTensor]]) -> Dict[str, torch.LongTensor]:
+            self.last_input = sample
+            return sample
+
+        def _update_pbar(sample: Union[BatchEncoding, Dict[str, torch.LongTensor]]) -> Dict[str, torch.LongTensor]:
+            self.progressbar[split].update(1)
+            return sample
+
         device_inputs_iter = map(_sample_to_device, inputs_iter)
 
         if add_oid:
@@ -284,8 +320,37 @@ class Objective(abc.ABC):
         if firstn is not None and firstn < self.dataset_length[split]:
             device_inputs_iter = itertools.islice(device_inputs_iter, firstn)
 
+        if self.remember_last_input:
+            device_inputs_iter = map(_remember_input, device_inputs_iter)
+
+        if show_progressbar:
+            device_inputs_iter = map(_update_pbar, device_inputs_iter)
+
         return TransformerAdaptationDataset(device_inputs_iter, self.dataset_length[split])
 
+    def compute_loss_on_last_sample(self) -> torch.FloatTensor:
+        """
+        This method aims to reproduce an error of calling the `objective.compatible_head_model`
+        on the `objective.last_input`. Useful for debugging new objective(s) in interactive (`-i`) mode.
+        """
+        if not self.remember_last_input:
+            raise ValueError("This objective does not remember its last output. "
+                             "For debugging, initialize the objective with `remember_last_input=True`.")
+
+        logger.warning("Reproducing loss computation on the last sample")
+        logger.warning("The last sample can be retrieved from `this_objective_instance.last_input`")
+        labels = self.last_input["labels"]
+
+        logger.warning("Computing model output")
+        model_inputs = {k: v for k, v in self.last_input.items() if k not in ("oid", "labels")}
+        logits = self.compatible_head_model(**model_inputs).logits
+
+        logger.warning("Computing loss")
+        loss = self._compute_loss(logits, labels, self.last_input)
+
+        logger.warning("Loss computation on the recent sample successful. Loss value: %s", loss.item())
+        return loss
+
     @abc.abstractmethod
     def _per_split_iterators(self, split: str) -> Union[Iterable[str], Tuple[Iterable[str], Iterable[str]]]:
         """
@@ -374,19 +439,26 @@ class UnsupervisedObjective(Objective, abc.ABC):
 
 
 class SupervisedObjective(UnsupervisedObjective, abc.ABC):
-
     labels_path: Optional[str] = None
     labels: Optional[List[str]] = None
 
     val_labels_path: Optional[str] = None
     val_labels: Optional[List[str]] = None
 
+    text_pair_path: Optional[str] = None
+    text_pair: Optional[List[str]] = None
+
+    val_text_pair_path: Optional[str] = None
+    val_text_pair: Optional[List[str]] = None
+
     labels_map: Dict[str, int] = {}
 
     def __init__(self,
                  *args,
                  labels_or_path: Union[str, List[str]],
                  val_labels_or_path: Optional[Union[str, List[str]]] = None,
+                 text_pair_or_path: Optional[Union[str, List[str]]] = None,
+                 val_text_pair_or_path: Optional[Union[str, List[str]]] = None,
                  **kwargs):
 
         if isinstance(labels_or_path, str):
@@ -400,6 +472,18 @@ class SupervisedObjective(UnsupervisedObjective, abc.ABC):
             else:
                 self.val_labels = val_labels_or_path
 
+        if text_pair_or_path is not None:
+            if isinstance(text_pair_or_path, str):
+                self.text_pair_path = text_pair_or_path
+            else:
+                self.text_pair = text_pair_or_path
+
+        if val_text_pair_or_path is not None:
+            if isinstance(val_text_pair_or_path, str):
+                self.val_text_pair_path = val_text_pair_or_path
+            else:
+                self.val_text_pair = val_text_pair_or_path
+
         # init will call register_compatible_head_model, which resolves num_labels for new head config from self.labels
         super().__init__(*args, **kwargs)
 
@@ -435,7 +519,44 @@ class SupervisedObjective(UnsupervisedObjective, abc.ABC):
         return super().register_compatible_head_model(lang_module, other_objective,
                                                       objective_args_for_head_config, preloaded_module)
 
-    def _per_split_iterators(self, split: str) -> Tuple[Iterable[str], Iterable[str]]:
+    def _get_inputs_iterator(self, split: str) -> Iterator[Union[BatchEncoding, Dict[str, torch.Tensor]]]:
+        """
+        Batches and encodes input texts and corresponding labels.
+        :param split: Selected data split. `train` or `eval`.
+        :return: Iterator over batch encodings.
+        """
+
+        collator = DataCollatorWithPadding(self.tokenizer, pad_to_multiple_of=8)
+        classifying_pairs = None
+
+        batch_features = []
+        for source_target_tuple in zip(*self._per_split_iterators(split)):
+            # check from the first sample
+            if classifying_pairs is None:
+                # if the input texts are tab-separated we will tokenize them as pairs
+                classifying_pairs = len(source_target_tuple) > 2
+                if classifying_pairs:
+                    assert len(source_target_tuple) == 3, "Expecting tuples of (source, source_pair, target) texts"
+            if classifying_pairs:
+                text, text_pair, label = source_target_tuple
+                out_sample = self.tokenizer(text, text_pair=text_pair, truncation=True)
+            else:
+                text, label = source_target_tuple
+                out_sample = self.tokenizer(text, truncation=True)
+
+            out_sample["label"] = torch.tensor(self.labels_map[label])
+
+            batch_features.append(out_sample)
+            if len(batch_features) == self.batch_size:
+                yield collator(batch_features)
+                batch_features = []
+
+        if batch_features:
+            # yield residual batch
+            yield collator(batch_features)
+
+    def _per_split_iterators(self, split: str) -> Union[Tuple[Iterable[str], Iterable[str]],
+                                                        Tuple[Iterable[str], Iterable[str], Iterable[str]]]:
         """
         Default inputs iterator for supervised objectives. Returns a pair of iterators, over input texts and labels.
         Not meant to be overriden when implementing custom data set. Instead choose to inherit either
@@ -449,8 +570,14 @@ class SupervisedObjective(UnsupervisedObjective, abc.ABC):
                 targets_iter = iter(self.labels)
             else:
                 targets_iter = AdaptationDataset.iter_text_file_per_line(self.labels_path)
-        elif split == "eval":
+            if self.text_pair is not None:
+                source_pairs_iter = iter(self.text_pair)
+            elif self.text_pair_path is not None:
+                source_pairs_iter = AdaptationDataset.iter_text_file_per_line(self.text_pair_path)
+            else:
+                source_pairs_iter = None
 
+        elif split == "eval":
             if self.val_labels is not None:
                 targets_iter = iter(self.val_labels)
             elif self.val_labels_path is not None:
@@ -459,7 +586,16 @@ class SupervisedObjective(UnsupervisedObjective, abc.ABC):
                 raise ValueError("Objective %s did not get any validation labels :( "
                                  "Hint: pass `AdaptationArgs(do_eval=False)` to avoid evaluation, "
                                  "or set Objective(val_labels) param." % self)
+
+            if self.val_text_pair is not None:
+                source_pairs_iter = iter(self.val_text_pair)
+            elif self.val_text_pair_path is not None:
+                source_pairs_iter = AdaptationDataset.iter_text_file_per_line(self.val_text_pair_path)
+            else:
+                source_pairs_iter = None
         else:
             raise ValueError("Unrecognized split: %s" % split)
-
-        return sources_iter, targets_iter
+        if source_pairs_iter is not None:
+            return sources_iter, source_pairs_iter, targets_iter
+        else:
+            return sources_iter, targets_iter
